/* This example demonstrates a bug in the app_timer library                     

The bug is a subtle race condition that can occur if during execution of timer_list_handler(),
which is executed in the context of the SWI0_EGU0_IRQn ISR, a higher priority ISR interrupts timer_list_handler 
and this ISR clears a timer then subsequently resets it.

The issue is that there is no protection around the timer operation queue in timer_list_handler.   
If a higher priority ISR executes between timer_list_handler performing the timer deletions and the timer insertions from the 
op list,  the stop operation by the higher priority ISR is not processed before the start.   When the start request is processed in 
list_insertions_handler the timer has not been removed from the active timer list, but the isRunning flag is false, thus
the test of this bit at app_timer.c:584 fails, continue is not executed and so the code proceeds to add the timer into the list.
This leads to a timer node having it's next pointer point to itself which subsequently causes an infinite loop next time
the timer list is traversed.

This sample demonstrates this by triggering a SWI_1 interrupt at the appropriate point in timer_list_handler simulating a real
interrupt occuring.  The handler for the SWI1 interrupt is configured to run at a higher priority than SWI0 interrupt.
The simulated interrupt is only triggered when global flag is set which is set by pressing button 1 on the PCA10040.

The code was compiled with IAR EWARM 7.60.1 with no optimisation Targetting the PCA10040

The original problem showed when using the soft device, but this example has stripped out all unrelated functionality

*/

#include <stdint.h>
#include <string.h>
#include "nordic_common.h"
#include "nrf.h"
#include "app_error.h"
#include "boards.h"
#include "app_timer.h"
#include "nrf_drv_gpiote.h"
#include "nrf_drv_clock.h"
#include "nrf_gpiote.h"
#include "app_util_platform.h"


#define APP_TIMER_PRESCALER              0                                           /**< Value of the RTC1 PRESCALER register. */
#define APP_TIMER_OP_QUEUE_SIZE          8                                           /**< Size of timer operation queues. */

#define SECS_TIMER_INTERVAL      APP_TIMER_TICKS(1000, APP_TIMER_PRESCALER)    /**Regular 1 second callback from timer module. */
#define TEST_TIMER_INTERVAL      APP_TIMER_TICKS(10000, APP_TIMER_PRESCALER)   /**<10 second 1 shot */



#define DEAD_BEEF                        0xDEADBEEF                                  /**< Value used as error code on stack dump, can be used to identify stack location on stack unwind. */


volatile bool simulateIrq;

APP_TIMER_DEF(m_test_timer_id);                        /**< test timer. */
APP_TIMER_DEF(m_secs_timer_id);                        /**< seconds timer. */


/**@brief Callback function for asserts in the SoftDevice.
 *
 * @details This function will be called in case of an assert in the SoftDevice.
 *
 * @warning This handler is an example only and does not fit a final product. You need to analyze
 *          how your product is supposed to react in case of Assert.
 * @warning On assert from the SoftDevice, the system can only recover on reset.
 *
 * @param[in] line_num   Line number of the failing ASSERT call.
 * @param[in] file_name  File name of the failing ASSERT call.
 */
void assert_nrf_callback(uint16_t line_num, const uint8_t * p_file_name)
{
    app_error_handler(DEAD_BEEF, line_num, p_file_name);
}




/**@brief Function for handling the seconds timer timeout.
 *
 *
 * @param[in] p_context  Pointer used for passing some arbitrary information (context) from the
 *                       app_start_timer() call to the timeout handler.
 */
static void seconds_timeout_handler(void * p_context)
{
  static int secs=0;
  secs++;
    UNUSED_PARAMETER(p_context);
    
}





/**@brief Function for handling the test timer timeout.
 *
 *
 * @param[in] p_context  Pointer used for passing some arbitrary information (context) from the
 *                       app_start_timer() call to the timeout handler.
 */
static void test_timeout_handler(void * p_context)
{
  static int ctr=0;
  ctr++;
    UNUSED_PARAMETER(p_context);
    
}





/**@brief Function for the Timer initialization.
 *
 * @details Initializes the timer module. This creates and starts application timers.
 */
static void timers_init(void)
{
    uint32_t err_code;

    // Initialize timer module.
    APP_TIMER_INIT(APP_TIMER_PRESCALER, APP_TIMER_OP_QUEUE_SIZE, false);

    // Create timers.
    err_code = app_timer_create(&m_test_timer_id,
//                                APP_TIMER_MODE_SINGLE_SHOT,
                                APP_TIMER_MODE_REPEATED,
                                test_timeout_handler);
    APP_ERROR_CHECK(err_code);

    err_code = app_timer_create(&m_secs_timer_id,
                                APP_TIMER_MODE_REPEATED,
                                seconds_timeout_handler);
    APP_ERROR_CHECK(err_code);

}





/**@brief Function for starting application timers.
 */
static void application_timers_start(void)
{
    uint32_t err_code;

    // Start application timers.
    err_code = app_timer_start(m_test_timer_id, TEST_TIMER_INTERVAL, NULL);
    APP_ERROR_CHECK(err_code);

    err_code = app_timer_start(m_secs_timer_id, SECS_TIMER_INTERVAL, NULL);
    APP_ERROR_CHECK(err_code);

}




void pin_event_handler(nrf_drv_gpiote_pin_t pin, nrf_gpiote_polarity_t action)
{
  simulateIrq = true;
}



void gpio_init()
{
  uint32_t err_code;
  if(!nrf_drv_gpiote_is_init())
  { 
    err_code = nrf_drv_gpiote_init();
    APP_ERROR_CHECK(err_code);
  }  
  
  nrf_drv_gpiote_in_config_t config = GPIOTE_CONFIG_IN_SENSE_HITOLO(true);
  config.pull = NRF_GPIO_PIN_PULLUP;
  err_code = nrf_drv_gpiote_in_init(BUTTON_1, &config, pin_event_handler);  
  APP_ERROR_CHECK(err_code);
// Event and corresponding interrupt are enabled.
  nrf_drv_gpiote_in_event_enable(BUTTON_1, true);  
}



/**@brief Function for handling the SWI interrupt.
 *
 * @details Performs all updates to the timer list.
 */
void SWI1_EGU1_IRQHandler(void)
{
  uint32_t err_code;
   // Stop test timer
  err_code = app_timer_stop(m_test_timer_id);
  APP_ERROR_CHECK(err_code);

  // Start test timer
  err_code = app_timer_start(m_test_timer_id, TEST_TIMER_INTERVAL, NULL);
  APP_ERROR_CHECK(err_code);    
}


/** @brief Function starting the internal LFCLK XTAL oscillator.
 */
static void lfclk_config(void)
{
    ret_code_t err_code = nrf_drv_clock_init();
    APP_ERROR_CHECK(err_code);

    nrf_drv_clock_lfclk_request(NULL);
}


/**@brief Function for application main entry.
 */
int main(void)
{
  

    // Initialize.
    lfclk_config();
    timers_init();
    gpio_init();
    application_timers_start();
    
    //Enable the SW interrupt
    NVIC_ClearPendingIRQ(SWI1_EGU1_IRQn);
    NVIC_SetPriority(SWI1_EGU1_IRQn, APP_IRQ_PRIORITY_HIGH);
    NVIC_EnableIRQ(SWI1_EGU1_IRQn);


    // Enter main loop.
    for (;;)
    {
       __WFE();
    }
}


